"""Collection of methods to substitute cols of open source component_to_resolve results
For most annotators we can re-use a common name.
Some make an exception and need special names derives, i.e. ClassifierDl/MultiClassifierDl/ etc.. which should be derived from nlu_ref.

"""
import logging

logger = logging.getLogger('nlu')


def partially_implemented_substitutor(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = nlu_identifier
    for col in cols:
        if 'results' in col:
            new_cols[col] = f'{new_base_name}_result'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif 'meta' in col:
            if 'confidence' in col:
                new_cols[col] = f"{new_base_name}_confidence"
            elif 'entity' in col:
                new_cols[col] = f"{new_base_name}_class"
            elif 'chunk' in col:
                new_cols[col] = f"{new_base_name}_origin_chunk"
            elif 'sentence' in col:
                new_cols[col] = f"{new_base_name}_origin_sentence"
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
                continue
        if '_embeddings' in col and f'{new_base_name}_embedding' not in new_cols.values():
            new_cols[col] = f'{new_base_name}_embedding'

    return new_cols


def substitute_ner_converter_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'entities' if nlu_identifier == 'UNIQUE' else f'entities_{nlu_identifier}' if 'entities' not in nlu_identifier else nlu_identifier
    for col in cols:
        if 'results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            new_cols[col] = f'{new_base_name}_embedding'
        elif 'meta' in col:
            if 'confidence' in col:
                new_cols[col] = f"{new_base_name}_confidence"
            elif 'entity' in col:
                new_cols[col] = f"{new_base_name}_class"
            elif 'sentence' in col:
                new_cols[col] = f"{new_base_name}_origin_sentence"
            elif 'chunk' in col:
                new_cols[col] = f"{new_base_name}_origin_chunk"
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_ner_dl_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'ner_iob' if nlu_identifier == 'UNIQUE' else f'ner_iob_{nlu_identifier}'
    for col in cols:
        if 'results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # always empty and irrelevant new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'confidence' in col:
                new_cols[col] = f"{new_base_name}_confidence"
            elif 'word' in col:
                continue  # is the same as token col, can be omitted
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_doc_assembler_cols(c, cols, nlu_identifier=True):
    """
    Doc assember is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'document'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # irrelevant  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                continue  # This field is irrelevant, since meta_sentence in document assembler is always 0
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_img_assembler_cols(c, cols, nlu_identifier=True):
    """
    Doc assember is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'image_origin' if nlu_identifier == 'UNIQUE' else f'image_origin{nlu_identifier}'
    for col in cols:
        if 'origin' in col:
            new_cols[col] = new_base_name
    return new_cols


def audio_assembler_cols(c, cols, nlu_identifier=True):
    """
    Sent detector is always unique
    """
    new_cols = {}
    new_base_name = 'audio_series' if nlu_identifier == 'UNIQUE' else f'audio_series_'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Sentence never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_wav2vec_cols(c, cols, nlu_identifier=True):
    """
    Sent detector is always unique
    """
    new_cols = {}
    new_base_name = 'text' if nlu_identifier == 'UNIQUE' else f'asr_text'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Sentence never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_sentence_detector_dl_cols(c, cols, nlu_identifier=True):
    """
    Sent detector is always unique
    """
    new_cols = {}
    new_base_name = 'sentence' if nlu_identifier == 'UNIQUE' else f'sentence_dl'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Sentence never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'sentence_sentence' in col:
                continue  # Seems like an irrelevant field, so drop
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_sentence_detector_pragmatic_cols(c, cols, nlu_identifier=True):
    """
    Sent detector is always unique
    """
    new_cols = {}
    new_base_name = 'sentence' if nlu_identifier == 'UNIQUE' else f'sentence_pragmatic'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Sentence never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'sentence_sentence' in col:
                continue  # Seems like an irrelevant field, so drop
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_tokenizer_cols(c, cols, nlu_identifier=True):
    """
    Tokenizer is always unique
    """
    new_cols = {}
    new_base_name = 'token'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_transformer_token_classifier_cols(c, cols, nlu_identifier=True):
    """
    Token classifier
    """
    new_cols = {}
    new_base_name = 'classified_token' if nlu_identifier == 'UNIQUE' else f'classified_token_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col and 'Some' not in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col: new_cols[
                col] = f'{new_base_name}_origin_sentence'
            if 'Some' in col:
                new_cols[
                    col] = f"'{new_base_name}_{col.split('Some(')[-1].split(')')[0]}_confidence"
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_coref_cols(c, cols, nlu_identifier=True):
    """
    |ORIGIN_REFERENCE | CO_REFERENCES|
    | Peter           | he , him, that dude |
    | Maria           | her, she, the lady |

    # {'head': 'ROOT', 'head.end': '-1', 'sentence': '0', 'head.sentence': '-1', 'head.begin': '-1'}

    """
    new_cols = {}
    new_base_name = 'coref' if nlu_identifier == 'UNIQUE' else f'coref_{nlu_identifier}'
    for col in cols:
        if '_result' in col:
            new_cols[col] = new_base_name
        elif '_sentence' in col:
            new_cols[col] = f'{new_base_name}_origin_sentence'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col and 'Some' not in col:
            continue
        elif '_types' in col:
            continue
        elif 'meta' in col:
            if 'head.sentence' in col:
                new_cols[col] = f'{new_base_name}_head_origin_sentence'
            elif 'head.end' in col:
                new_cols[col] = f'{new_base_name}_head_end'
            elif 'head.begin' in col:
                new_cols[col] = f'{new_base_name}_head_begin'
            elif '_head' in col:
                new_cols[col] = f'{new_base_name}_head'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_seq_bert_classifier_cols(c, cols, nlu_identifier=True):
    """
    Seq classifier
    """
    new_cols = {}

    # Remove sequence from nlu_identifier because we use it in the base name anyways
    if '_sequence' in nlu_identifier:
        nlu_identifier = nlu_identifier.replace('_sequence', '')
    new_base_name = 'classified_sequence' if nlu_identifier == 'UNIQUE' else f'classified_sequence_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col and 'Some' not in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            if '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            if 'Some' in col:
                new_cols[col] = f"'{new_base_name}_{col.split('Some(')[-1].split(')')[0]}_confidence"
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols


def substitute_span_classifier_cols(c, cols, nlu_identifier=True):
    """
    Seq classifier
    """
    new_cols = {}
    # Remove sequence from nlu_identifier because we use it in the base name anyways
    # if '_sequence' in nlu_identifier:
    #     nlu_identifier = nlu_identifier.replace('_sequence', '')
    new_base_name = 'answer' if nlu_identifier == 'UNIQUE' else f'{nlu_identifier}_answer'
    for col in cols:
        if 'span_result' in col:
            new_cols[col] = f'{new_base_name}'
        if 'span_score' in col:
            new_cols[col] = f'{new_base_name}_confidence'

        elif 'span_start_score' in col:
            new_cols[col] = f'{new_base_name}_start_confidence'
        elif 'span_end_score' in col:
            new_cols[col] = f'{new_base_name}_end_confidence'
        elif 'start' in col and not 'score' in col:
            new_cols[col] = f'{new_base_name}_start'
        elif 'end' in col and not 'score' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif 'sentence' in col:
            new_cols[col] = f'{new_base_name}_sentence'

    return new_cols


def substitute_multi_doc_span_assembler_cols(c, cols, nlu_identifier=True):
    """
    Seq classifier
    """
    new_cols = {}
    # Remove sequence from nlu_identifier because we use it in the base name anyways
    for col in cols:
        if 'question' in col and not 'context' in col:
            new_cols[col] = f'question'
        elif 'context' in col:
            new_cols[col] = f'context'
    return new_cols


def substitute_tapas_qa_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = f'tapas_qa_{nlu_identifier}' if 'tapas_qa_' not in nlu_identifier else nlu_identifier
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}_answer'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'question' in col:
                new_cols[col] = f'{new_base_name}_origin_question'
            elif 'aggregation' in col:
                new_cols[col] = f'{new_base_name}_aggregation'
            elif 'cell_positions' in col:
                new_cols[col] = f'{new_base_name}_cell_positions'
            elif 'cell_scores' in col:
                new_cols[col] = f'{new_base_name}_cell_scores'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_word_embed_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'word_embedding_{nlu_identifier}' if 'word_embedding_' not in nlu_identifier else nlu_identifier  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            continue  # new_cols[col] = new_base_name can be omitted for word_embeddings, maps to the origin token, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif 'OOV' in col:
                new_cols[col] = f'{new_base_name}_is_OOV'
            elif 'isWordStart' in col:
                new_cols[col] = f'{new_base_name}_is_word_start'
            elif 'pieceId' in col:
                new_cols[col] = f'{new_base_name}_piece_id'
            elif '_token' in col:
                continue  # Can be omitted, is the same as _result, just maps to origin_token
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        elif '_embeddings' in col:
            new_cols[col] = new_base_name  # stores the embeds and represents basically the main result

    return new_cols


def substitute_sent_embed_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    # new_base_name = f'sentence_embedding_{nlu_identifier}'  # if nlu_identifier else f'document_{nlu_identifier}'
    new_base_name = f'sentence_embedding_{nlu_identifier}' if 'sentence_embedding' not in nlu_identifier else nlu_identifier  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            continue  # new_cols[col] = new_base_name can be omitted for word_embeddings, maps to the origin token, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'OOV' in col:
                new_cols[col] = f'{new_base_name}_is_OOV'
            elif 'isWordStart' in col:
                new_cols[col] = f'{new_base_name}_is_word_start'
            elif 'pieceId' in col:
                new_cols[col] = f'{new_base_name}_piece_id'
            elif '_token' in col:
                continue  # Can be omited, is the same as _result, just maps to origin_token
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        elif '_embeddings' in col:
            new_cols[col] = new_base_name  # stores the embeds and represents basically the main result
        elif '_sentence' in col and 'meta' in col:
            new_cols[col] = f'{new_base_name}_origin_sentence'

    return new_cols


def substitute_chunk_embed_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for chunk Embeddings. For Word_Embeddings, some name will be infered, and chunk_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'chunk_embedding_{nlu_identifier}' if 'chunk_embedding_' not in nlu_identifier else nlu_identifier  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            continue  # new_cols[col] = new_base_name can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif 'isWordStart' in col:
                new_cols[col] = f'{new_base_name}_is_word_start'
            elif 'pieceId' in col:
                new_cols[col] = f'{new_base_name}_piece_id'
            elif '_token' in col:
                continue  # Can be omited, is the same as _result, just maps to origin_token
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        elif '_embeddings' in col:
            new_cols[col] = new_base_name  # stores the embeds and represents basically the main result

    return new_cols


def substitute_classifier_dl_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'{nlu_identifier}'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[
                col] = new_base_name  # can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            old_base_name = f'meta_{c.out_types[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif metadata in ['confidence', '_confidence']:
                new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            # else:
            #     new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for os_components={os_components}')

    return new_cols


def substitute_ngram_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. ngram will be the new base name
    """
    new_cols = {}
    new_base_name = f'ngram'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols


def substitute_labled_dependency_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy labeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'labeled_dependency'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols


def substitute_un_labled_dependency_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy unlabeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'unlabeled_dependency'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_head.begin' in col:
                new_cols[col] = f'{new_base_name}_head_begin'
            elif 'head.end' in col:
                new_cols[col] = f'{new_base_name}_head_end'
            elif 'head' in col:
                new_cols[col] = f'{new_base_name}_head'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols


def substitute_pos_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy unlabeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'pos'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_word' in col:
                continue  # can be omitted, is jsut the token
            elif 'confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_norm_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for normalized,  <norm> will be new base col name
    """
    new_cols = {}
    new_base_name = f'norm'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_doc_norm_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for normalized,  <norm> will be new base col name
    """
    new_cols = {}
    new_base_name = f'doc_norm'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_finisher_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name finisher. For now keeps original name
    """
    return dict(zip(cols,cols))

def substitute_spell_context_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for normalized,  <spell> will be new base col namem
    1 spell checker is assumed per component_list for now
    """
    new_cols = {}
    new_base_name = f'spell' if nlu_identifier == 'UNIQUE' else f'spell_dl'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col: new_cols[col] = f'{new_base_name}_origin_sentence'
            if '_cost' in col:
                new_cols[col] = f'{new_base_name}_cost'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_spell_symm_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for sym,  <spell> will be new base col name
    1 spell checker is assumed per component_list for now
    """
    new_cols = {}
    new_base_name = f'spell' if nlu_identifier == 'UNIQUE' else f'spell_sym'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col: new_cols[col] = f'{new_base_name}_origin_sentence'
            if '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_spell_norvig_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for spell,  <spell> will be new base col name
    1 spell checker is assumed per component_list for now
    """
    new_cols = {}
    new_base_name = f'spell' if nlu_identifier == 'UNIQUE' else f'spell_norvig'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col: new_cols[col] = f'{new_base_name}_origin_sentence'
            if '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_word_seg_cols(c, cols, nlu_identifier=True):
    """
    Word_seg is always unique
    """
    new_cols = {}
    new_base_name = 'words_seg'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_stem_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'stem'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_lem_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'lem'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_stopwords_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'stopword_less'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_chunk_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_pos'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_YAKE_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'keywords'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col: new_cols[
                col] = f'{new_base_name}_origin_sentence'
            if '_score' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_marian_cols(c, cols, nlu_identifier=True):
    """
    rename cols with base name either <translated> or if not unique <translated_<lang>>
    """
    new_cols = {}
    new_base_name = 'translated' if nlu_identifier == 'UNIQUE' else f'translated_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_gpt2_cols(c, cols, nlu_identifier=True):
    """
    rename cols with base name either <gpt> or if not unique <generated_identifier>
    """
    new_cols = {}
    new_base_name = 'generated' if nlu_identifier == 'UNIQUE' else f'generated_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_T5_cols(c, cols, nlu_identifier=True):
    """
    rename cols with base name either <t5> or if not unique <t5_<task>>
    """
    new_cols = {}
    new_base_name = 't5' if nlu_identifier == 'UNIQUE' else f't5_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols

def substitute_summarizer_cols(c, cols, nlu_identifier=True):
    """
    rename cols with base name either <t5> or if not unique <t5_<task>>
    """
    new_cols = {}
    new_base_name = 'summary' if nlu_identifier == 'UNIQUE' else f'summary_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_sentiment_vivk_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'sentiment' if nlu_identifier == 'UNIQUE' else f'sentiment_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_sentiment_dl_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'sentiment' if nlu_identifier == 'UNIQUE' else f'sentiment_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            elif '_negative' in col:
                new_cols[col] = f'{new_base_name}_negative'
            elif '_positive' in col:
                new_cols[col] = f'{new_base_name}_positive'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_multi_classifier_dl_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'{nlu_identifier}'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[
                col] = new_base_name  # can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            old_base_name = f'meta_{c.out_types[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif metadata == 'confidence':
                new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else:
                new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for os_components={os_components}')

    return new_cols


def substitute_date_match_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_date'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_regex_match_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_regex'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_text_match_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_text'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


## Trainable
def substitute_classifier_dl_approach_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'trained_classifier'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[
                col] = new_base_name  # can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            old_base_name = f'meta_{c.out_types[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif metadata == 'confidence':
                new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else:
                new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for os_components={os_components}')
    return new_cols


def substitute_sentiment_vivk_approach_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'trained_sentiment'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_sentiment_dl_approach_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'trained_sentiment'
    for col in cols:
        if '_results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            elif '_negative' in col:
                new_cols[col] = f'{new_base_name}_negative'
            elif '_positive' in col:
                new_cols[col] = f'{new_base_name}_positive'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_multi_classifier_dl_approach_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'trained_multi_classifier'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[
                col] = new_base_name  # can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            old_base_name = f'meta_{c.out_types[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif metadata == 'confidence':
                new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else:
                new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for os_components={os_components}')

    return new_cols


def substitute_ner_dl_approach_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'trained_ner_iob'
    for col in cols:
        if 'results' in col:
            new_cols[col] = new_base_name
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col:
            continue  # always empty and irrelevant new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col:
            continue  # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'confidence' in col:
                new_cols[col] = f"{new_base_name}_confidence"
            elif 'word' in col:
                continue  # is the same as token col, can be omitted
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_pos_approach_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy unlabeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'trained_pos'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            if '_sentence' in col:
                new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_word' in col:
                continue  # can be omitted, is jsut the token
            elif 'confidence' in col:
                new_cols[col] = f'{new_base_name}_confidence'
            else:
                logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_doc2chunk_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Doc2chunk
    """
    new_cols = {}
    new_base_name = f'doc2chunk'  # if nlu_identifier else f'document_{nlu_identifier}'
    for col in cols:
        if '_results' in col:
            new_cols[col] = f'{new_base_name}'
        elif '_beginnings' in col:
            new_cols[col] = f'{new_base_name}_begin'
        elif '_endings' in col:
            new_cols[col] = f'{new_base_name}_end'
        elif '_types' in col:
            continue  #
        elif '_embeddings' in col:
            continue  #
        elif 'meta' in col:
            logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols




class NERDLHealthcare:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import MedicalNerModel
        return MedicalNerModel.pretrained(name='ner_dl_bert', lang='en') \
            .setInputCols(["sentence", "token", "word_embeddings"]) \
            .setOutputCol("ner") \
            .setIncludeConfidence(True)



    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import MedicalNerModel
        return MedicalNerModel.pretrained(name,language,bucket) \
            .setInputCols(["sentence", "token", "word_embeddings"]) \
            .setOutputCol("ner") \
            .setIncludeConfidence(True)

    @staticmethod
    def get_default_trainable_model():
        from sparknlp_jsl.annotator import MedicalNerApproach
        return MedicalNerApproach() \
            .setInputCols(["sentence", "token", "word_embeddings"]) \
            .setLabelColumn("y") \
            .setOutputCol("ner") \
            .setMaxEpochs(2) \
            .setVerbose(0) \
            .setIncludeConfidence(True)


from sparknlp_jsl.annotator import *
from sparknlp.base import *
from sparknlp_display import *

from nlu import NLP_FEATURES
from nlu.universe.feature_node_ids import NLP_HC_NODE_IDS, NLP_NODE_IDS


class VizUtilsHC():
    """Utils for interfacing with the Spark-NLP-Display lib - licensed Viz"""
    HTML_WRAPPER = """<div class="scroll entities" style="overflow-x: auto; border: 1px solid #e6e9ef; border-radius: 0.25rem; padding: 1rem; margin-bottom: 2.5rem; white-space:pre-wrap">{}</div>"""

    @staticmethod
    def infer_viz_licensed(pipe) -> str:
        """For a given NLUPipeline with licensed components, infers which visualizations are applicable. """
        # we go in reverse, which makes NER always take lowest priority and NER feeder annotators have higher priority
        for c in pipe.components[::-1]:
            if isinstance(c.model, TypedDependencyParserModel):
                return 'dep'
            if isinstance(c.model, (SentenceEntityResolverModel)):
                return 'resolution'
            if isinstance(c.model, (RelationExtractionDLModel, RelationExtractionDLModel)):
                return 'relation'
            if isinstance(c.model, (AssertionDLModel, AssertionLogRegModel)):
                return 'assert'
            if isinstance(c.model, (NerConverter, NerConverterInternal)):
                return 'ner'

    @staticmethod
    def viz_ner(anno_res, pipe, labels=[], viz_colors={}, is_databricks_env=False, write_to_streamlit=False,
                ner_col=None):
        """Infer columns required for ner viz and then viz it.
        viz_colors :  set label colors by specifying hex codes , i.e. viz_colors =  {'LOC':'#800080', 'PER':'#77b5fe'}
        labels : only allow these labels to be displayed. (default: [] - all labels will be displayed)
        """
        document_col, entities_col = VizUtilsHC.infer_ner_dependencies(pipe)
        if ner_col:
            entities_col = ner_col
        ner_vis = NerVisualizer()
        if len(viz_colors) > 0: ner_vis.set_label_colors(viz_colors)
        if write_to_streamlit:
            import streamlit as st
            HTML = ner_vis.display(anno_res, label_col=entities_col, document_col=document_col, labels=labels,
                                   return_html=True)
            CSS, HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)

        elif not is_databricks_env:
            ner_vis.display(anno_res, label_col=entities_col, document_col=document_col, labels=labels)
        else:
            return ner_vis.display(anno_res, label_col=entities_col, document_col=document_col, labels=labels,
                                   return_html=True)

    @staticmethod
    def infer_ner_dependencies(pipe):
        """Finds entities and doc cols for ner viz"""
        doc_component = None
        entities_component = None
        for c in pipe.components:

            if NLP_FEATURES.NAMED_ENTITY_CONVERTED in c.out_types:
                entities_component = c
            if isinstance(c.model, DocumentAssembler):
                doc_component = c

        document_col = doc_component.spark_output_column_names[0]
        entities_col = entities_component.spark_output_column_names[0]
        return document_col, entities_col

    @staticmethod
    def viz_dep(anno_res, pipe, is_databricks_env, write_to_streamlit=False, user_dep_untyped_col=None,
                user_dep_typed_col=None, user_pos_col=None):
        """Viz dep result"""
        pos_col, dep_typ_col, dep_untyp_col = VizUtilsHC.infer_dep_dependencies(pipe)
        if user_dep_untyped_col:
            dep_untyp_col = user_dep_untyped_col

        if user_dep_typed_col:
            dep_typ_col = user_dep_typed_col
        if user_pos_col:
            pos_col = user_pos_col

        dependency_vis = DependencyParserVisualizer()
        if write_to_streamlit:
            import streamlit as st
            SVG = dependency_vis.display(anno_res, pos_col=pos_col, dependency_col=dep_untyp_col,
                                         dependency_type_col=dep_typ_col, return_html=True)
            # st.markdown(SVG, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(SVG), unsafe_allow_html=True)
        elif not is_databricks_env:
            dependency_vis.display(anno_res, pos_col=pos_col, dependency_col=dep_untyp_col,
                                   dependency_type_col=dep_typ_col)
        else:
            return dependency_vis.display(anno_res, pos_col=pos_col, dependency_col=dep_untyp_col,
                                          dependency_type_col=dep_typ_col, return_html=True)

    @staticmethod
    def infer_dep_dependencies(pipe):
        """Finds entities,pos,dep_typed,dep_untyped and  doc cols for dep viz viz"""
        # doc_component      = None
        pos_component = None
        dep_untyped_component = None
        dep_typed_component = None
        for c in pipe.components:
            if isinstance(c.model, PerceptronModel):              pos_component = c
            if isinstance(c.model, TypedDependencyParserModel):   dep_typed_component = c
            if isinstance(c.model, DependencyParserModel):        dep_untyped_component = c

        pos_col = pos_component.spark_output_column_names[0]
        dep_typ_col = dep_typed_component.spark_output_column_names[0]
        dep_untyp_col = dep_untyped_component.spark_output_column_names[0]
        return pos_col, dep_typ_col, dep_untyp_col

    @staticmethod
    def viz_resolution(anno_res, pipe, viz_colors={}, is_databricks_env=False, write_to_streamlit=False,
                       user_ner_col=None, user_resolution_col=None, ):
        """Viz dep result. Set label colors by specifying hex codes, i.e. viz_colors={'TREATMENT':'#800080', 'PROBLEM':'#77b5fe'} """
        entities_col, resolution_col, doc_col = VizUtilsHC.infer_resolution_dependencies(pipe)
        er_vis = EntityResolverVisualizer()
        if len(viz_colors) > 0: er_vis.set_label_colors(viz_colors)

        if user_ner_col:
            entities_col = user_ner_col
        if user_resolution_col:
            resolution_col = user_resolution_col

        if write_to_streamlit:
            import streamlit as st
            HTML = er_vis.display(anno_res, label_col=entities_col, resolution_col=resolution_col, document_col=doc_col,
                                  return_html=True)
            CSS, HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)


        elif not is_databricks_env:
            er_vis.display(anno_res, label_col=entities_col, resolution_col=resolution_col, document_col=doc_col)
        else:
            return er_vis.display(anno_res, label_col=entities_col, resolution_col=resolution_col, document_col=doc_col,
                                  return_html=True)

    @staticmethod
    def infer_resolution_dependencies(pipe):
        """Finds entities_col,resolution_col,doc_col cols for resolution viz viz"""
        entities_component, resolution_component, doc_component = None, None, None
        for c in pipe.components:
            if c.name == NLP_NODE_IDS.DOCUMENT_ASSEMBLER:
                doc_component = c
            if NLP_FEATURES.NAMED_ENTITY_CONVERTED in c.out_types:
                entities_component = c

            if c.name == NLP_HC_NODE_IDS.SENTENCE_ENTITY_RESOLVER:
                resolution_component = c
        entities_col = entities_component.spark_output_column_names[0]
        resolution_col = resolution_component.spark_output_column_names[0]
        doc_col = doc_component.spark_output_column_names[0]
        return entities_col, resolution_col, doc_col

    @staticmethod
    def viz_relation(anno_res, pipe, is_databricks_env, write_to_streamlit=False, user_relation_col=None):
        """Viz relation result. Set label colors by specifying hex codes, i.e. viz_colors={'TREATMENT':'#800080', 'PROBLEM':'#77b5fe'} """
        relation_col, document_col = VizUtilsHC.infer_relation_dependencies(pipe)
        if user_relation_col:
            relation_col = user_relation_col

        re_vis = RelationExtractionVisualizer()
        if write_to_streamlit:
            import streamlit as st
            HTML = re_vis.display(anno_res, relation_col=relation_col, document_col=document_col, show_relations=True,
                                  return_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)

        if not is_databricks_env:
            re_vis.display(anno_res, relation_col=relation_col, document_col=document_col, show_relations=True)
        else:
            return re_vis.display(anno_res, relation_col=relation_col, document_col=document_col, show_relations=True,
                                  return_html=True)

    @staticmethod
    def infer_relation_dependencies(pipe):
        """Finds relation_col,document_col  cols for relation viz viz"""
        relation_component, doc_component = None, None
        for c in pipe.components:
            if isinstance(c.model, DocumentAssembler):              doc_component = c
            if isinstance(c.model, (RelationExtractionDLModel, RelationExtractionModel)):   relation_component = c
        relation_col = relation_component.spark_output_column_names[0]
        document_col = doc_component.spark_output_column_names[0]
        return relation_col, document_col

    @staticmethod
    def viz_assertion(anno_res, pipe, viz_colors={}, is_databricks_env=False, write_to_streamlit=False,
                      user_ner_col=None, user_assertion_col=None
                      ):
        """Viz relation result. Set label colors by specifying hex codes, i.e. viz_colors={'TREATMENT':'#008080', 'problem':'#800080'} """
        entities_col, assertion_col, doc_col = VizUtilsHC.infer_assertion_dependencies(pipe)
        assertion_vis = AssertionVisualizer()
        if user_ner_col:
            entities_col = user_ner_col
        if user_assertion_col:
            assertion_col = assertion_col

        if len(viz_colors) > 0: assertion_vis.set_label_colors(viz_colors)
        if write_to_streamlit:
            import streamlit as st
            HTML = assertion_vis.display(anno_res, label_col=entities_col, assertion_col=assertion_col,
                                         document_col=doc_col, return_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            CSS, HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)
        elif not is_databricks_env:
            assertion_vis.display(anno_res, label_col=entities_col, assertion_col=assertion_col, document_col=doc_col)
        else:
            return assertion_vis.display(anno_res, label_col=entities_col, assertion_col=assertion_col,
                                         document_col=doc_col, return_html=True)

    @staticmethod
    def infer_assertion_dependencies(pipe):
        """Finds relation_col,document_col  cols for relation viz viz"""
        entities_component, assert_component, doc_component = None, None, None
        for c in pipe.components:
            if c.name == NLP_NODE_IDS.DOCUMENT_ASSEMBLER:
                doc_component = c
            if c.name in [NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL, NLP_NODE_IDS.NER_CONVERTER,
                          NLP_NODE_IDS.PARTIAL_NerConverterInternalModel]:
                entities_component = c
            if c.name in [NLP_HC_NODE_IDS.ASSERTION_DL, NLP_HC_NODE_IDS.ASSERTION_LOG_REG]:
                assert_component = c
        entities_col = entities_component.spark_output_column_names[0]
        assertion_col = assert_component.spark_output_column_names[0]
        doc_col = doc_component.spark_output_column_names[0]
        return entities_col, assertion_col, doc_col



from sparknlp.annotator import BertSentenceEmbeddings


class BertSentence:
    @staticmethod
    def get_default_model():
        return BertSentenceEmbeddings.pretrained() \
        .setInputCols("sentence") \
        .setOutputCol("sentence_embeddings")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return BertSentenceEmbeddings.pretrained(name,language,bucket) \
            .setInputCols('sentence') \
            .setOutputCol("sentence_embeddings")






import sparknlp

class RegexMatcher:
    @staticmethod
    def get_default_model():
        return sparknlp.annotator.RegexMatcher()\
            .setInputCols("sentence") \
            .setOutputCol("regex_entity") \
            

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return RegexMatcher.get_default_model()
        # sparknlp.annotator.TextMatcherModel.pretrained(name,language,bucket) \
        #     .setInputCols("document") \
        #     .setOutputCol("entity") \





from sparknlp.annotator import *

class T5:
    @staticmethod
    def get_default_model():
        return T5Transformer.pretrained() \
        .setInputCols("document") \
        .setOutputCol("T5")

    @staticmethod
    def get_pretrained_model(name, language,bucket=None):
        return T5Transformer.pretrained(name, language,bucket) \
            .setInputCols("document") \
            .setOutputCol("T5")










class ContextParser:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import ContextualParserApproach
        return ContextualParserApproach() \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("context_entity")\
            .setCaseSensitive(False) \
            .setContextMatch(False) \
            .setPrefixAndSuffixMatch(False)
    @staticmethod
    def get_default_trainable_model():
        from sparknlp_jsl.annotator import ContextualParserApproach
        return ContextualParserApproach() \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("context_entity") \
            .setCaseSensitive(False) \
            .setContextMatch(False) \
            .setPrefixAndSuffixMatch(False)



from sparknlp.base import TableAssembler

class SparkNlpTableAssembler:
    @staticmethod
    def get_default_model():
        return TableAssembler() \
            .setInputCols('raw_table') \
            .setOutputCol("assembled_table")



from sparknlp.annotator import RoBertaForTokenClassification

class TokenRoBerta:
    @staticmethod
    def get_default_model():
        return RoBertaForTokenClassification.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("ner")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return RoBertaForTokenClassification.pretrained(name, language,bucket) \
            .setInputCols("sentence", "token") \
            .setOutputCol("ner")






from sparknlp import AudioAssembler


class AudioAssembler_:
    @staticmethod
    def get_default_model():
        return AudioAssembler() \
            .setInputCol("audio_content") \
            .setOutputCol("audio_assembler")





"""
for sentence resolvers
SENTENCE EMBEDS must be generated from CHUNK2DOC!!!!!!!!!!!!!
chunk2doc = Chunk2Doc().setInputCols("entities@modifier").setOutputCol("chunk2doc")
"""
class SentenceResolver:
    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        from sparknlp_jsl.annotator import SentenceEntityResolverModel
        return SentenceEntityResolverModel.pretrained(name, language,bucket) \
            .setInputCols(['entities',"sentence_embeddings"]) \
            .setDistanceFunction("COSINE") \
            .setNeighbours(3) \
            .setOutputCol("sentence_resolution")

    @staticmethod
    def get_default_trainable_model():
        from sparknlp_jsl.annotator import SentenceEntityResolverApproach
        return SentenceEntityResolverApproach() \
            .setNeighbours(25) \
            .setThreshold(1000) \
            .setInputCols(['entities',"sentence_embeddings"]) \
            .setNormalizedCol("_y") \
            .setLabelCol("y") \
            .setOutputCol('sentence_resolution') \
            .setDistanceFunction("EUCLIDIAN") \
            .setCaseSensitive(False)





"""
This module contains methods that will be applied inside of the apply call on the freshly convertet
Spark DF that is now as list of dicts.

These methods are meant to be applied in the internal calls of the metadata extraction.
They expect dictionaries which represent the metadata field extracted from Spark NLP annotators.


"""
import numpy as np


def meta_extract_language_classifier_max_confidence(row, configs):
    ''' Extract the language classificationw ith highest confidence and drop the others '''
    # # todo Get the best, but what about TOP K!  conditional sentence extraction and mroe docs
    # unpack all confidences to float and set 'sentence' key value to -1 so it does not affect finding the highest cnfidence
    unpack_dict_values = lambda x: -1 if 'sentence' in x[0] else float(x[1][0])
    l = list(map(unpack_dict_values, row.items()))
    m = np.argmax(l)
    k = list(row.keys())[m]

    return {k + '_confidence': row[k][0]}  # remoe [0] for list return


def zipp(l): return zip(*l)  # unpack during list comprehension not supported in Python, need this workaround for now


def extract_maximum_confidence(row, configs):
    ''' Extract the maximum confidence from any classifier with N classes.
    A classifier with N classes, has N confidences in it's metadata by default, which is too much data usually.
    This extractor gets the highest confidence from the array of confidences.
    This method assumes all keys in metadata corrospond to confidences, except the `sentence` key, which maps to a sentence ID
    key schema is 'meta_' + configs.output_col_prefix + '_confidence'

    Parameters
    -------------
    configs : SparkNLPExtractorConfig
    if configs.get_sentence_origin is True, the sentence origin column will be kept, otherwise dropped.
    row : dict
        i.e. looks like{'meta_category_sentence': ['0'],'meta_category_surprise': ['0.0050183665'],'meta_category_sadness': ['8.706827E-5'],'meta_category_joy': ['0.9947379'],'meta_category_fear': ['1.5667251E-4']}
    Returns
    ------------
    dict
      if configs.get_sentence_origin True  {'meta_sentiment_dl_sentence': ['0', '1'], 'meta_sentiment_dl_confidence': [0.9366506, 0.9366506]}
      else {'meta_sentiment_dl_confidence': [0.9366506, 0.9366506]}
    '''
    meta_sent_key = 'meta_' + configs.output_col_prefix + '_sentence'
    fl = lambda \
            k: False if 'sentence' in k else True  # every key  that has not the sub string sentence in it is considerd a confidence key
    confidences_keys = list(filter(fl, row.keys()))

    if configs.pop_meta_list:
        return {
            **{'meta_' + configs.output_col_prefix + '_confidence': max([float(row[k][0]) for k in confidences_keys])},
            **({'meta_' + configs.output_col_prefix + '_sentence': row[
                meta_sent_key]} if configs.get_sentence_origin else {})
        }
    else:
        if len(confidences_keys) == 1:
            return {
                **{'meta_' + configs.output_col_prefix + '_confidence': max([float(row[k]) for k in confidences_keys])},
                **({'meta_' + configs.output_col_prefix + '_sentence': row[
                    meta_sent_key]} if configs.get_sentence_origin else {})
            }
        else:
            return {
                **{'meta_' + configs.output_col_prefix + '_confidence': [max(z) for z in zipp(
                    list(map(float, row[k])) for k in confidences_keys)]},
                **({'meta_' + configs.output_col_prefix + '_sentence': row[
                    meta_sent_key]} if configs.get_sentence_origin else {})
            }


def unpack_HPO_codes(row, k):
    # Case: HPO
    UMLS_codes = []
    ORPHA_CODES = []
    MSH_CODES = []
    SNOMED_CODES = []
    OMIM_CODES = []

    for resolution in row[k]:
        k_candidates = resolution.split(':::')
        for candidate in k_candidates:
            # There is 0 to 5 alt terminologies, 1 per HPO_codes. If one is missing, we have to append None
            alt_terminologies = candidate.split('||')
            UMLS_ok, ORPHA_ok, MSH_ok, SNOMED_ok, OMIM_ok = False, False, False, False, False
            for alt in alt_terminologies:
                if 'UMLS' in alt:
                    UMLS_codes.append(alt)
                    UMLS_ok = True
                elif 'ORPHA' in alt:
                    ORPHA_CODES.append(alt)
                    ORPHA_ok = True
                elif 'MSH' in alt:
                    MSH_CODES.append(alt)
                    MSH_ok = True
                elif 'SNOMED' in alt:
                    SNOMED_CODES.append(alt)
                    SNOMED_ok = True
                elif 'OMIM' in alt:
                    OMIM_CODES.append(alt)
                    OMIM_ok = True
            # Detect which of the 0 to 5 alt terminologies are missing and add None for them
            if not UMLS_ok:
                UMLS_codes.append(None)
            if not ORPHA_ok:
                ORPHA_CODES.append(None)
            if not MSH_ok:
                MSH_CODES.append(None)
            if not SNOMED_ok:
                SNOMED_CODES.append(None)
            if not OMIM_ok:
                OMIM_CODES.append(None)
    return UMLS_codes, ORPHA_CODES, MSH_CODES, SNOMED_CODES, OMIM_CODES,
    # Write into dict


def extract_resolver_all_k_subfields_splitted(row, configs):
    ''' Extract all metadata fields for sentence resolver annotators and splits all _k_ fields on ::: , relevant for all_k_result, all_k_resolutions, al_k_distances, all_k_cosine_distances
    pop_meta_list should be true, if outputlevel of pipe is the same as the resolver component we are extracting here for
    ::: for icd
    || in HCC splits billable/status/code, always aligns. in HPO splits resolutions in alterantive terminologies, not always aligns
    ::: splits Mappings of resolutions, i.e. len(split(:::)) == number_of_entities  == number of resolutions
    Special case HPO :MeSH/SNOMED/UMLS/ORPHA/OMIM


unpack_resolutions = lambda x : x.split(':::')
unpack_k_aux_label = lambda x : x.split('||')
# each of the k resolutions as K aux label.
# for HPO, each aux label, can have 0 to 5 extra terminolgies
unpack_terms = lambda x: x.split('')

unpacked = list(map(lambda x: list(map(unpack_k_aux_label, unpack_resolutions(x))), row[k] ))

for key, g in itertools.groupby(unpacked, lambda x : x.split(':')[0]):
    print(key,list(g))

unpacked[2]

    :UMLS
    '''
    # todo del AUX label col for CODES
    HPO_CODES = ['UMLS', 'ORPHA', 'MSH', 'SNOMED', 'OMIM']
    prefix = 'meta_' + configs.output_col_prefix + '_'
    res = {}
    for k in row.keys():
        if '_k_' in k:
            if '||' in row[k][0]:

                if any(x in row[k][0] for x in HPO_CODES):
                    # Case : HPO
                    res[prefix + 'k_UMLS_codes'], \
                    res[prefix + 'k_ORPHA_codes'], \
                    res[prefix + 'k_MESH_codes'], \
                    res[prefix + 'k_SNOMED_codes'], \
                    res[prefix + 'k_OMIM_codes'] = unpack_HPO_codes(row, k)

                else:
                    # CASE : billable code handling
                    f = lambda s: list(map(lambda x: x.split("||"), s.split(':::')))
                    # Triple assignment so we unpack properly
                    res[prefix + 'billable'], \
                    res[prefix + 'hcc_status'], \
                    res[prefix + 'hcc_code'] = zip(*map(lambda x: zip(*x), map(f, row[k])))
                    # Casting from tuple to list or we get problems during pd explode
                    h = lambda z: list(map(lambda r: list(r), z))# [0]
                    res[prefix + 'billable'] = h(res[prefix + 'billable'])
                    res[prefix + 'hcc_status'] = h(res[prefix + 'hcc_status'])
                    res[prefix + 'hcc_code'] = h(res[prefix + 'hcc_code'])

            elif ':::' in row[k][0]:
                # CASE : General code handling
                res[prefix+k.replace('results', 'codes')] = list(map(lambda x: x.split(':::'), row[k]))# [0]
        else:
            # Any other metadata field hadling
            res[k] = row[k]
    return res


def extract_chunk_mapper_relation_data(row, configs):
    ''' Splits all_relations field on ::: to create an array ,


    uses row.relation as prefix


    '''

    prefix = 'meta_' + configs.output_col_prefix + '_'
    for k in row.keys():
        if 'chunk_all_relations' in k:
            row[k] = [s.split(':::') for s in row[k]]
    return row



def extract_coreference_data(row_metadata,row_results, configs):
    ''' Splits all_relations field on ::: to create an array ,
    | Text                                                                    |heads         | Co-References    | Heads_sentence| Coref_sentence| coref_head_begin | coref_head_end|
    |John told Mary he would like to borrow a book from her, after his lunch  |[John, Marry]  | [he,his], [her] | [0,0]         | [0,0],[0]     | [0,0], [10]      | [3,3], [13]


    |Text     | heads  | Co-Refernces
    | John    | ROOT   |   [he,his]
    | told    | /      |  /
    | Marry   |  ROOT | [her]
    | he      | JOHN  | /
    | likes   | /     |  /
    | her     | MARRY |/



    |ORIGIN_REFERENCE | CO_REFERENCES|
    | Peter           | he , him, that dude |
    | Maria           | her, she, the lady |


    '''
    head_to_coref = {}
    prefix = 'meta_' + configs.output_col_prefix + '_'

    # for (k_meta,v_meta), (k_result,v_result) in zip(row_metadata.items(), row_results.items()):
    #     if

    raise NotImplemented('Not implemented')



"""
This module contains methods that will be applied inside of the apply call on the freshly convertet
Spark DF that is now as list of dicts.

These methods are meant to be applied in the internal calls of the metadata extraction.
They expect dictionaries which represent the metadata field extracted from Spark NLP annotators.


"""
from functools import reduce

import numpy as np
import pandas as pd
import pyspark
from pyspark.sql import Row as PysparkRow
from sparknlp.annotation import Annotation

from nlu.pipe.extractors.extractor_base_data_classes import *


def extract_light_pipe_rows(df):
    """Extract Annotations from Light Pipeline into same represenation as other extractors in thos module"""
    ff = lambda row: list(map(f, row)) if isinstance(row, List) else row
    f = lambda anno: dict(
                # annotatorType=anno.annotator_type,
                          begin=anno.begin, end=anno.end,
                          result=anno.result,
                          metadata=anno.metadata,
                          embeddings=anno.embeddings if isinstance(anno, List) else []) \
        if isinstance(anno, Annotation) else anno
    return df.applymap(ff)


def extract_pyspark_rows(r: pd.Series, ) -> pd.Series:
    """ Convert pyspark.sql.Row[Annotation] to List(Dict[str,str]) objects. Except for key=metadata in dict,
    this element in the Dict which is [str,Dict[str,str]] Checks if elements are of type list and whether they contain
    Pyspark Rows. If PysparkRow, call .asDict() on every row element to generate the dicts First method that runs
    after toPandas() call
    """
    if isinstance(r, str):
        return r
    elif isinstance(r, list):
        if len(r) == 0:
            return r
        elif isinstance(r[0], PysparkRow):
            pyspark_row_to_list = lambda l: l.asDict()
            return list(map(pyspark_row_to_list, r))
    return r


def extract_pyarrow_rows(r: pd.Series, ) -> pd.Series:
    """ Convert pyspark.sql.Row[Annotation] to List(Dict[str,str]) objects. Except for key=metadata in dict,
    this element in the Dict which is [str,Dict[str,str]]
    Checks if elements are of type list and whether they contain Pyspark Rows.
     If PysparkRow, call .asDict() on every row element to generate the dicts
    First method that runs after toPandas() call
    """
    if isinstance(r, str):
        return r
    elif isinstance(r, np.ndarray):
        if len(r) == 0:
            return r
        elif isinstance(r[0], dict) and 'annotatorType' in r[0].keys():
            r[0]['metadata'] = dict(r[0]['metadata'])
            return r
    return r


def extract_base_sparkocr_features(row: pd.Series, configs: SparkOCRExtractorConfig) -> dict:
    ###### OCR EXTRACTOR
    # for now only text recognizer outputs fetched
    if configs.name == 'default text recognizer config':
        if configs.get_text:
            return {'text': row}
    # Check for primitive type here and return
    if 'visual_classifier' in configs.name:
        # Either Label or Confidence
        if isinstance(row, str):
            return {'visual_classifier_label': row}
        else:
            return {'visual_classifier_confidence': row}

    else:
        # # OCR unpackers (TODO WIP)
        # unpack_text = lambda x: unpack_dict_list(x, 'text')
        # # unpack_image = lambda x : unpack_dict_list(x, 'TODO') # is data?
        # unpack_image_origin = lambda x: unpack_dict_list(x, 'origin')
        # unpack_image_height = lambda x: unpack_dict_list(x, 'height')
        # unpack_image_width = lambda x: unpack_dict_list(x, 'width')
        # unpack_image_n_channels = lambda x: unpack_dict_list(x, 'nChannels')
        # unpack_image_mode = lambda x: unpack_dict_list(x, 'mode')
        # unpack_image_resolution = lambda x: unpack_dict_list(x, 'resolution')
        # unpack_image_data = lambda x: unpack_dict_list(x, 'data')
        # # unpack_path = lambda x : unpack_dict_list(x, 'TODO')
        # unpack_modification_time = lambda x: unpack_dict_list(x, 'TODO')
        # unpack_length = lambda x: unpack_dict_list(x, 'TODO')
        # unpack_page_num = lambda x: unpack_dict_list(x, 'TODO')
        # unpack_confidence = lambda x: unpack_dict_list(x, 'TODO')
        # unpack_exception = lambda x: unpack_dict_list(x, 'TODO')
        # unpack_img_positions = lambda x: unpack_dict_list(x, 'TODO')
        # if configs.get_image:
        #     pass
        # if configs.get_image_origin:
        #     pass
        # if configs.get_image_height:
        #     pass
        # if configs.get_image_width:
        #     pass
        # if configs.get_image_n_channels:
        #     pass
        # if configs.get_image_mode:
        #     pass
        # if configs.get_image_resolution:
        #     pass
        # if configs.get_image_data:
        #     pass
        # if configs.get_path:
        #     pass
        # if configs.get_modification_time:
        #     pass
        # if configs.get_length:
        #     pass
        # if configs.get_page_num:
        #     pass
        # if configs.get_confidence:
        #     pass
        # if configs.get_exception:
        #     pass
        # if configs.get_img_positions:
        #     pass
        # return {**beginnings, **endings, **results, **annotator_types, **embeddings}  # Merge dicts OCR output

        return {}


def extract_base_sparknlp_features(row: pd.Series, configs: SparkNLPExtractorConfig) -> dict:
    """
    Extract base features common in all saprk NLP annotators
    Begin/End/Embedding/Metadata/Result, except for the blacklisted features
    Expects a list with Token Annotator Outputs from extract_pyspark_rows() , i.e
    Setting pop to true for a certain field will return only the first element of that fields list of elements. Useful if that field always has exactly 1 result, like many classifirs
    [{'annotatorType': 'token',
    'begin': 0,
    'embeddings': [],
    'end': 4,
    'metadata': {'sentence': '0'},
    'result': 'Hello'
    }]
    row = pyspark.row

    or

    [
      {'annotatorType': 'language',
  'begin': 0,
  'embeddings': [],
  'end': 57,
  'metadata': {'bg': '0.0',
   'sentence': '0',
   'sl': '5.2462015E-24',
   'sv': '2.5977007E-25'},
  'result': 'en'}
  ]

    returns a DICT
    """

    unpack_dict_list = lambda d, k: d[k]
    unpack_begin = lambda x: unpack_dict_list(x, 'begin')
    unpack_end = lambda x: unpack_dict_list(x, 'end')
    unpack_annotator_type = lambda x: unpack_dict_list(x, 'annotatorType')
    unpack_result = lambda x: unpack_dict_list(x, 'result')
    unpack_embeddings = lambda x: unpack_dict_list(x, 'embeddings')
    unpack_origin = lambda x: unpack_dict_list(x, 'origin')

    # Either extract list of anno results and put them in a dict with corrosponding key name or return empty dict {} for easy merge in return
    annotator_types = {configs.output_col_prefix + '_types': list(
        map(unpack_annotator_type, row))} if configs.get_annotator_type else {}
    # Same logic as above, but we check wether to pop or not and either evaluate the map result with list() or just next()

    origins = {configs.output_col_prefix + '_origin': next(map(unpack_origin, row))} if configs.get_origin else {}

    if configs.pop_result_list:
        results = {configs.output_col_prefix + '_results': next(map(unpack_result, row))} if configs.get_result else {}
    else:
        results = {configs.output_col_prefix + '_results': list(map(unpack_result, row))} if configs.get_result else {}
    if configs.pop_begin_list:
        beginnings = {configs.output_col_prefix + '_beginnings': next(
            map(unpack_begin, row))} if configs.get_begin or configs.get_positions else {}
    else:
        beginnings = {configs.output_col_prefix + '_beginnings': list(
            map(unpack_begin, row))} if configs.get_begin or configs.get_positions else {}

    if configs.pop_end_list:
        endings = {configs.output_col_prefix + '_endings': next(
            map(unpack_end, row))} if configs.get_end or configs.get_positions else {}
    else:
        endings = {configs.output_col_prefix + '_endings': list(
            map(unpack_end, row))} if configs.get_end or configs.get_positions else {}

    if configs.pop_embeds_list:
        embeddings = {
            configs.output_col_prefix + '_embeddings': next(map(unpack_embeddings, row))} if configs.get_embeds else {}
    else:
        embeddings = {
            configs.output_col_prefix + '_embeddings': list(map(unpack_embeddings, row))} if configs.get_embeds else {}

    return {**beginnings, **endings, **results, **annotator_types, **embeddings, **origins}  # Merge dicts NLP output


def extract_sparknlp_metadata(row: pd.Series, configs: SparkNLPExtractorConfig) -> dict:
    """
    Extract base features common in all saprk NLP annotators
    Begin/End/Embedding/Metadata/Result, except for the blacklisted features
    Expects a list with Token Annotator Outputs, i.e.
    Can either use a WHITE_LISTE or BLACK_LIST or get ALL metadata
    For WHITE_LIST != [], only metadata keys/values will be kepts, for which the keys are contained in the white list
    For WHITE_LIST == [] AND BLACK_LIST !=, all metadata key/values will be returned, which are not on the black list.
    If  WHITE_LIST is not [] the BLACK_LIST will be ignored.

    returns one DICT which will be merged into pd.Serise by the extractor calling this exctractor for .apply() in pythonify
    """
    if len(row) == 0: return {}
    unpack_dict_list = lambda d, k: d[k]
    # extract list of metadata dictionaries (all dict should have same keys)
    unpack_metadata_to_dict_list = lambda x: unpack_dict_list(x, 'metadata')

    metadatas_dict_list = list(map(unpack_metadata_to_dict_list, row))
    # extract keys, which should all be equal in all rows

    if configs.get_full_meta:
        keys_in_metadata = list(metadatas_dict_list[0].keys()) if len(metadatas_dict_list) > 0 else []
    elif len(configs.meta_white_list) != 0:
        keys_in_metadata = [k for k in metadatas_dict_list[0].keys() if k in configs.meta_white_list]
    elif len(configs.meta_black_list) != 0:
        keys_in_metadata = [k for k in metadatas_dict_list[0].keys() if k not in configs.meta_black_list]
    else:
        keys_in_metadata = []

    # dectorate lambda with key to extract, equalt to def decorate_f(key): return lambda x,y :  x+ [y[key]]
    # For a list of dicts which all have the same keys, will return a list of all the values for one key in all the dicts
    if configs.pop_meta_list:
        f = lambda key: metadatas_dict_list[0][key]
        metadata_scalars = list(map(f, keys_in_metadata))
        result = dict(
            zip(map(lambda x: 'meta_' + configs.output_col_prefix + '_' + x, keys_in_metadata), metadata_scalars))
        return result
    extract_val_from_dic_list_to_list = lambda key: lambda x, y: x + [y[key]] if key in y else x + [None]
    # List of lambda expression, on for each Key to be extracted. (TODO balcklisting?)
    dict_value_extractors = list(map(extract_val_from_dic_list_to_list, keys_in_metadata))
    # reduce list of dicts with same struct and a common key to a list of values for thay key. Leveraging closuer for meta_dict_list
    reduce_dict_list_to_values = lambda t: reduce(t, metadatas_dict_list, [])
    # list of lists, where each list is corrosponding to all values in the previous dict list
    meta_values_list = list(map(reduce_dict_list_to_values, dict_value_extractors))
    # add prefix to key and zip with values for final dict result
    result = dict(
        zip(list(map(lambda x: 'meta_' + configs.output_col_prefix + '_' + x, keys_in_metadata)), meta_values_list))
    return result


def extract_master(row: pd.Series, configs: SparkNLPExtractorConfig) -> pd.Series:
    """
    Re-Usable base extractor for simple Annotators like Document/Token/etc..?
    extract_universal/?/Better name?
    row = a list or Spark-NLP annotations as dictionary
    """
    if isinstance(row, pyspark.sql.Row) and len(row) == 0:
        return pd.Series({})
    if isinstance(configs, SparkOCRExtractorConfig):
        base_annos = extract_base_sparkocr_features(row, configs)
    else:
        if isinstance(configs, FinisherExtractorConfig):
            # 1. if normal finisher col, just return val -> {finisher_col: [values]}
            # 2. if meta finisher col, return grouped meta dict ->{m1: [v1,v2,..], m2: [v1,v2,..], ...}
            # 3. double check with other anno behaviour
            if configs.is_meta_field:
                return pd.Series(
                    {
                        configs.source_col_name: row
                    })
            else:
                return pd.Series(
                    {
                        configs.source_col_name: row
                    })

        else:
            base_annos = extract_base_sparknlp_features(row, configs)

    # Get Metadata
    all_metas = extract_sparknlp_metadata(row, configs) if configs.get_meta or configs.get_full_meta else {}

    # Apply custom extractor methods
    if configs.meta_data_extractor.name != '':
        if configs.meta_data_extractor.extractor_with_result_method:
            all_metas = configs.meta_data_extractor.extractor_with_result_method(all_metas, base_annos, configs)
        else:
            all_metas = configs.meta_data_extractor.extractor_method(all_metas, configs)

    # Apply Finishers on metadata/additional fields
    return pd.Series(
        {
            **base_annos,
            **all_metas
        })




def apply_extractors_and_merge(df, anno_2_ex_config, keep_stranger_features, stranger_features):
    """ apply extract_master on all fields with corrosponding configs after converting Pyspark Rows to List[Dict]
    and merge them to a final DF (1 to 1 mapping still)
    df  The Df we want to apply the extractors on
    columns_to_extractor_map Map column names to extractor configs. Columns which are not in these keys will be ignored These configs will be passed to master_extractor for every column
    """
    # keep df and ex_resolver in closure and apply base extractor with configs for each col
    extractor = lambda c: df[c].apply(extract_master, configs=anno_2_ex_config[c])
    keep_strangers = lambda c: df[c]

    # merged_extraction_df
    # apply the extract_master together with it's configs to every column and geenrate a list of output DF's, one per Spark NLP COL
    # TODO handle MULTI-COL-OUTPUT. If Anno has multi cols, then we either needs multiple keys in anno_2_ex or use something besides
    # anno_2_ex_config.keys() here because it will only apply to one of the extracted rows..(?)

    # Apply each Anno Extractor to the corrosponding generated col.
    # If no Extractor defined for a col and it is not a stranger feature, it will be dropped here
    return pd.concat(
        list(map(extractor, anno_2_ex_config.keys())) +
        list(map(keep_strangers, stranger_features)) if keep_stranger_features else [],
        axis=1)


def pad_same_level_cols(row):
    """We must ensure that the cols which are going to be exploded have all the same amount of elements.
    To ensure this, we apply this methods on the cols we wish to explode. It ensures, they have all the same
    length and can be exploded eronous free
    """
    max_len = 0
    lens = {}
    for c in row.index:
        if isinstance(row[c], list):
            lens[c] = len(row[c])
            if lens[c] > max_len: max_len = lens[c]
        else:
            lens[c] = 1
            row[c] = [row[c]]

    for c, length in lens.items():
        if length < max_len:
            row[c] += [np.nan] * (max_len - length)
    return row


def zip_and_explode(df: pd.DataFrame, cols_to_explode: List[str]) -> pd.DataFrame:
    """
    Returns a new dataframe, where columns in cols_to_explode should have all array elements.
    :param df: Dataframe to explode columns on. Each column in cols_to_explode should be of type array
    :param cols_to_explode: list of columns to explode
    :return: new dataframe, where each array element of a row in cols_to_explode is in a new row.
            For exploding Rows where the lists are same length,
            lists will be padded to length of the longest list in that row (sub-levels)
            Elements of columns which are not in cols_to_explode, will be in lists
    """
    # Check cols we want to explode actually exist, if no data extracted cols can be missing
    missing = []
    for col in cols_to_explode:
        if col not in df.columns:
            missing.append(col)
    for miss in missing:
        cols_to_explode.remove(miss)
    # Drop duplicate cols
    df = df.loc[:, ~df.columns.duplicated()]
    if len(cols_to_explode) > 0:
        # We must pad all cols we want to explode to the same length because pandas limitation.
        # Spark API does not require this since it handles cols with not same length by creating nan. We do it ourselves here manually
        df[cols_to_explode] = df[cols_to_explode].apply(pad_same_level_cols, axis=1)
        return pd.concat([df.explode(c)[c] for c in cols_to_explode] + [df.drop(cols_to_explode, axis=1)], axis=1)
    else:
        # No padding
        return pd.concat([df.drop(cols_to_explode, axis=1)], axis=1)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

REQUIRED_PKGS = [
    'spark-nlp>=5.0.2',
    'numpy',
    'pyarrow>=0.16.0',
    'pandas>=1.3.5',
    'dataclasses'
]

setup(

    name='nlu',

    version='5.1.5rc2',

    description='John Snow Labs NLU provides state of the art algorithms for NLP&NLU with 20000+ of pretrained models in 200+ languages. It enables swift and simple development and research with its powerful Pythonic and Keras inspired API. It is powerd by John Snow Labs powerful Spark NLP library.',

    long_description=long_description,
    install_requires=REQUIRED_PKGS,

    long_description_content_type='text/markdown',

    url='http://nlu.johnsnowlabs.com',
    author='John Snow Labs',
    author_email='christian@johnsnowlabs.com',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='NLP spark development NLU ',

    packages=find_packages(exclude=['test*', 'tmp*']),
    include_package_data=True
)




import logging

from nlu.pipe.nlu_component import NluComponent
from nlu.pipe.utils.resolution import uid_to_storage_ref as uid2storageref

logger = logging.getLogger('nlu')

"""Storage Ref logic operations and utils"""


class StorageRefUtils:
    @staticmethod
    def has_storage_ref(component: NluComponent):
        """Storage ref is either on the model_anno_obj or nlu component_to_resolve defined """
        return component.has_storage_ref

    @staticmethod
    def extract_storage_ref(component: NluComponent):
        """Extract storage ref from either a NLU component_to_resolve or NLP Annotator. First checks if annotator has storage
        ref, otherwise check NLU attribute """
        if StorageRefUtils.has_storage_ref(component):
            return StorageRefUtils.nlp_extract_storage_ref_nlp_model(component)
        else:
            raise ValueError(
                f'Tried to extract storage ref from component_to_resolve which has no storageref ! Component = {component}')

    @staticmethod
    def fallback_storage_ref_resolutions(storage_ref):
        """
        For every storage ref result, we check if its storage ref is defined as its UID and if a fallback storageref
        is available. If available, alternative is returned, otherwise original
        """
        if storage_ref in uid2storageref.mappings.keys():
            return uid2storageref.mappings[storage_ref]
        else:
            return storage_ref

    @staticmethod
    def has_component_storage_ref_or_anno_storage_ref(component: NluComponent):
        """Storage ref is either on the model_anno_obj or nlu component_to_resolve defined """
        return component.has_storage_ref

    @staticmethod
    def nlp_component_has_storage_ref(model):
        """Check if a storage ref is defined on the Spark NLP Annotator model_anno_obj"""
        for k, _ in model.extractParamMap().items():
            if k.name == 'storageRef':
                return True
        return False

    @staticmethod
    def extract_storage_ref_from_component(component):
        """Extract storage ref from a NLU component_to_resolve which embellished a Spark NLP Annotator"""
        if StorageRefUtils.nlu_component_has_storage_ref(component):
            return component.info.storage_ref
        elif StorageRefUtils.nlp_component_has_storage_ref(component):
            return StorageRefUtils.nlp_extract_storage_ref_nlp_model(component)
        else:
            return ''

    @staticmethod
    def nlu_extract_storage_ref_nlp_model(component):
        """Extract storage ref from a NLU component_to_resolve which embellished a Spark NLP Annotator"""
        return component.model.extractParamMap()[component.model.getParam('storageRef')]

    @staticmethod
    def nlu_component_has_storage_ref(component):
        """Check if a storage ref is defined on the Spark NLP Annotator embellished by the NLU Component"""
        if hasattr(component.info, 'storage_ref'):
            return True
        return False

    @staticmethod
    def nlp_extract_storage_ref_nlp_model(component: NluComponent):
        """Extract storage ref from a NLU component_to_resolve which embellished a Spark NLP Annotator"""
        # Embedding Converters don't have storage ref attribute on class, but NLU component_to_resolve has attribute for it
        params = list(component.model.extractParamMap().keys())
        for p in params:
            if p.name == 'storageRef':
                storage_ref = component.model.extractParamMap()[component.model.getParam('storageRef')]
                if not storage_ref:
                    # For untrained components storage ref will be none
                    return ''
                else:
                    return storage_ref
        if not component.storage_ref:
            return ''
        return component.storage_ref



"""Get data into JVM for prediction and out again as Spark Dataframe"""
import glob
import logging
import os
from typing import List
import pyspark
import numpy as np
import pandas as pd

logger = logging.getLogger('nlu')


class OcrDataConversionUtils:
    @staticmethod
    def validate_OCR_compatible_inputs(data):
        """Validate for input data that it contains a path pointing to file or jsl_folder"""
        if isinstance(data, List):
            return OcrDataConversionUtils.check_iterable_paths_are_valid(data)
        if isinstance(data, str):
            return OcrDataConversionUtils.check_iterable_paths_are_valid([data])
        if isinstance(data, pd.DataFrame):
            return 'path' in data.columns
        if isinstance(data, pd.Series):
            return 'path' in data.name
        if isinstance(data, pyspark.sql.dataframe.DataFrame):
            return 'path' in data.columns
        if isinstance(data, np.ndarray):
            return OcrDataConversionUtils.check_iterable_paths_are_valid(data)

    @staticmethod
    def check_iterable_paths_are_valid(iterable_paths):
        """Validate for iterable data input if all elements point to file or jsl_folder"""
        paths_validness = []
        for p in iterable_paths:
            if os.path.isdir(p) or os.path.isfile(p):
                paths_validness.append(True)
            else:
                print(f'Warning : Invalid path for jsl_folder or file in input. Could validate path.\n'
                      f'NLU will try and ignore this issue but you might run into errors.\n'
                      f'Please make sure all paths are valid\n')
                print(f'For path = {p}')
                paths_validness.append(False)
        return all(paths_validness)

    @staticmethod
    def check_all_paths_point_to_accepted_file_type(paths, file_types):
        """Validate that all paths point to a file type defined by file_types"""
        pass

    @staticmethod
    def glob_files_of_accepted_type(paths, file_types):
        """Get all paths which point to correct file types from iterable paths which can contain file and jsl_folder paths
        1. paths point to a file which is suffixed with one of the accepted file_types, i.e. path/to/file.type
        2. path points to a jsl_folder, in this case jsl_folder is recurisvely searched for valid files and accepted paths will be in return result
        """
        accepted_file_paths = []
        for p in paths:
            for t in file_types:
                t = t.lower()
                if os.path.isfile(p):
                    if p.lower().split('.')[-1] == t:
                        accepted_file_paths.append(p)
                elif os.path.isdir(p):
                    accepted_file_paths += glob.glob(p + f'/*.{t.upper()}', recursive=True) + glob.glob(p + f'/*.{t}', recursive=True)
                else:
                    print(f"Invalid path = {p} pointing neither to file or jsl_folder on this machine")
        return accepted_file_paths

    @staticmethod
    def extract_iterable_paths_from_data(data):
        """Extract an iterable object containing paths from input data"""
        if isinstance(data, List):
            return data
        if isinstance(data, str):
            return [data]
        if isinstance(data, pd.DataFrame):
            return list(data['path'].values)
        if isinstance(data, pd.Series):
            return list(data.values)
        if isinstance(data, pyspark.sql.dataframe.DataFrame):
            return [p['path'] for p in data.select('path').collect()]
        if isinstance(data, np.ndarray):
            return list(data)

    @staticmethod
    def get_accepted_ocr_file_types(pipe):
        """Get all file typtes/suffixes that can be processed by the pipeline"""
        accepted_files = []
        for c in pipe.components:
            if c.applicable_file_types:
                accepted_files += c.applicable_file_types
        return list(set(accepted_files))



import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from typing import List, Optional
import streamlit as st
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker


class SentenceEmbeddingManifoldStreamlitBlock():
    @staticmethod
    def viz_streamlit_sentence_embed_manifold(
            pipe,  # nlu component_list
            default_texts: List[str] = (
            "Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
            title: Optional[str] = "Lower dimensional Manifold visualization for Sentence embeddings",
            sub_title: Optional[
                str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Sentence Embeddings` to `1-D`, `2-D` and `3-D` ",
            write_raw_pandas: bool = False,
            default_algos_to_apply: List[str] = ("TSNE", "PCA"),
            # ,'LLE','Spectral Embedding','MDS','ISOMAP','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
            target_dimensions: List[int] = (1, 2, 3),
            show_algo_select: bool = True,
            show_embed_select: bool = True,
            show_color_select: bool = True,
            MAX_DISPLAY_NUM: int = 200000,
            display_embed_information: bool = True,
            set_wide_layout_CSS: bool = True,
            num_cols: int = 3,
            model_select_position: str = 'side',  # side or main
            key: str = "NLU_streamlit",
            additional_classifiers_for_coloring: List[str] = ['sentiment.imdb'],
            generate_code_sample: bool = False,
            show_infos: bool = True,
            show_logo: bool = True,
            n_jobs: Optional[int] = 3,  # False
    ):
        from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
        StreamlitVizTracker.footer_displayed = False

        try:
            import plotly.express as px
            from sklearn.metrics.pairwise import distance_metrics
        except:
            st.error(
                "You need the sklearn and plotly package in your Python environment installed for similarity visualizations. Run <pip install sklearn plotly>")
        # if len(default_texts) > MAX_DISPLAY_NUM : default_texts = default_texts[:MAX_DISPLAY_NUM]
        if show_logo: StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS: _set_block_container_style()
        if title: st.header(title)
        if sub_title: st.subheader(sub_title)
        # if show_logo :VizUtilsStreamlitOS.show_logo()

        # VizUtilsStreamlitOS.loaded_word_embeding_pipes = []

        data = st.text_area('Enter N texts, seperated by new lines to visualize Sentence Embeddings for ',
                            default_texts)
        # detect_sentence = False # TODO ITNEGRATE PARAM
        output_level = 'document'  # if not detect_sentence else 'sentence'
        classifier_cols = []
        original_text = nlu.load('tokenize').predict(data.split("\n"), output_level=output_level)[output_level].values
        original_text = original_text
        original_text = original_text[original_text != '']
        original_text = original_text[~pd.isna(original_text)]

        text_col = output_level
        embed_algos_to_load = []
        class_algos_to_load = []
        new_embed_pipes = []
        new_class_pipes = []
        e_coms = StreamlitUtilsOS.find_all_embed_components(pipe)

        if show_algo_select:
            exp = st.expander("Select additional manifold and dimension reduction techniques to apply")

            algos = exp.multiselect(
                "Reduce embedding dimensionality to something visualizable",
                options=(
                "TSNE", "ISOMAP", 'LLE', 'Spectral Embedding', 'MDS', 'PCA', 'SVD aka LSA', 'DictionaryLearning',
                'FactorAnalysis', 'FastICA', 'KernelPCA', 'LatentDirichletAllocation'),
                default=default_algos_to_apply, )

            emb_components_usable = [e for e in Discoverer.get_components('embed', True, include_aliases=True) if
                                     'chunk' not in e and 'sentence' in e]
            # Todo, multi-classifiers excluded
            classifier_components_usable = [e for e in Discoverer.get_components('classify', True, include_aliases=True)
                                            if 'xx' not in e and 'toxic' not in e and 'e2e' not in e]
            # Storage Ref extraction
            loaded_embed_nlu_refs, loaded_storage_refs = StreamlitUtilsOS.extract_all_sentence_storage_refs_or_nlu_refs(
                e_coms)
            loaded_classifier_nlu_refs = additional_classifiers_for_coloring  # + all classifier NLU_refs?

            # Get loaded Embed NLU Refs
            for embed_pipe in StreamlitVizTracker.loaded_sentence_embeding_pipes:
                if embed_pipe != pipe: loaded_embed_nlu_refs.append(embed_pipe.nlu_ref)
            loaded_embed_nlu_refs = list(set(loaded_embed_nlu_refs))

            # Get loaded Classifier NLU Refs
            for embed_pipe in StreamlitVizTracker.loaded_document_classifier_pipes:
                if embed_pipe != pipe: loaded_classifier_nlu_refs.append(embed_pipe.nlu_ref)
            loaded_classifier_nlu_refs = list(set(loaded_classifier_nlu_refs))

            # fix default selector
            for l in loaded_embed_nlu_refs:
                if l not in emb_components_usable: emb_components_usable.append(l)

            # fix default selector
            for l in loaded_classifier_nlu_refs:
                if l not in classifier_components_usable: classifier_components_usable.append(l)

            emb_components_usable.sort()
            loaded_embed_nlu_refs.sort()
            classifier_components_usable.sort()
            loaded_classifier_nlu_refs.sort()
            if model_select_position == 'side':
                embed_algo_selection = st.sidebar.multiselect(
                    "Pick additional Sentence Embeddings for the Dimension Reduction", options=emb_components_usable,
                    default=loaded_embed_nlu_refs, key=key)
                embed_algo_selection = [embed_algo_selection[-1]]

                exp = st.expander("Pick additional Classifiers")
                class_algo_selection = exp.multiselect("Pick additional Classifiers to load for coloring points",
                                                       options=classifier_components_usable,
                                                       default=loaded_classifier_nlu_refs, key=key)
                class_algo_selection = [class_algo_selection[-1]]

            else:
                exp = st.expander("Pick additional Sentence Embeddings")
                embed_algo_selection = exp.multiselect(
                    "Pick additional Sentence Embeddings for the Dimension Reduction", options=emb_components_usable,
                    default=loaded_embed_nlu_refs, key=key)
                embed_algo_selection = [embed_algo_selection[-1]]

                exp = st.expander("Pick additional Classifiers")
                class_algo_selection = exp.multiselect("Pick additional Classifiers to load for coloring points",
                                                       options=classifier_components_usable,
                                                       default=loaded_classifier_nlu_refs, key=key)
                class_algo_selection = [class_algo_selection[-1]]

            embed_algos_to_load = list(set(embed_algo_selection) - set(loaded_embed_nlu_refs))
            class_algos_to_load = list(set(class_algo_selection) - set(loaded_classifier_nlu_refs))

        for embedder in embed_algos_to_load: new_embed_pipes.append(nlu.load(embedder))
        for classifier in class_algos_to_load: new_class_pipes.append(nlu.load(classifier))

        StreamlitVizTracker.loaded_sentence_embeding_pipes += new_embed_pipes
        StreamlitVizTracker.loaded_document_classifier_pipes += new_class_pipes
        if pipe not in StreamlitVizTracker.loaded_sentence_embeding_pipes: StreamlitVizTracker.loaded_sentence_embeding_pipes.append(
            pipe)

        for nlu_ref in additional_classifiers_for_coloring:  # TODO REMVOVE< INTEGRATE INTO THE AUT LOAD THING REDUNDAND
            already_loaded = False
            for embed_pipe in StreamlitVizTracker.loaded_document_classifier_pipes:
                if embed_pipe.nlu_ref == nlu_ref: already_loaded = True
            if not already_loaded:
                already_loaded = True
                StreamlitVizTracker.loaded_document_classifier_pipes.append(nlu.load(nlu_ref))

        col_index = 0
        cols = st.columns(num_cols)

        data = original_text.copy()
        # Get classifier predictions
        classifier_cols = []
        for class_pipe in StreamlitVizTracker.loaded_document_classifier_pipes:
            data = class_pipe.predict(data, output_level=output_level, multithread=False)
            classifier_cols += StreamlitUtilsOS.get_classifier_cols(class_pipe)
            data['text'] = original_text
            # drop embeds of classifiers because bad conversion
            for c in data.columns:
                if 'embedding' in c: data.drop(c, inplace=True, axis=1)

        data['text'] = original_text
        if show_color_select:
            if model_select_position == 'side':
                feature_to_color_by = st.sidebar.selectbox('Pick a feature to color points in manifold by ',
                                                           classifier_cols, 0)
            else:
                feature_to_color_by = st.selectbox('Feature to color plots by ', classifier_cols, 0)

        def are_cols_full():
            return col_index == num_cols

        for embed_pipe in StreamlitVizTracker.loaded_sentence_embeding_pipes:
            predictions = embed_pipe.predict(data, output_level=output_level, multithread=False).dropna()
            e_col = StreamlitUtilsOS.find_embed_col(predictions)
            e_com = StreamlitUtilsOS.find_embed_component(embed_pipe)
            e_com_storage_ref = StorageRefUtils.extract_storage_ref(e_com)
            emb = predictions[e_col]
            mat = np.array([x for x in emb])
            for algo in algos:
                # Only pos values for latent Dirchlet
                if algo == 'LatentDirichletAllocation': mat = np.square(mat)
                if len(mat.shape) > 2: mat = mat.reshape(len(emb), mat.shape[-1])
                hover_data = classifier_cols + ['text']
                # calc reduced dimensionality with every algo
                if 1 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo, 1, n_jobs).fit_transform(mat)
                    x = low_dim_data[:, 0]
                    y = np.zeros(low_dim_data[:, 0].shape)
                    predictions['text'] = original_text
                    tsne_df = pd.DataFrame({**{'x': x, 'y': y},
                                            **{k: predictions[k] for k in classifier_cols},
                                            **{'text': original_text}
                                            })
                    fig = px.scatter(tsne_df, x="x", y="y", color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Sentence-Embeddings =`{e_com_storage_ref}`, Manifold-Algo =`{algo}` for `D=1`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig, key=key)
                    col_index += 1
                    if are_cols_full():
                        cols = st.columns(num_cols)
                        col_index = 0
                if 2 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo, 2, n_jobs).fit_transform(mat)
                    x = low_dim_data[:, 0]
                    y = low_dim_data[:, 1]
                    tsne_df = pd.DataFrame({**{'x': x, 'y': y},
                                            **{k: predictions[k] for k in classifier_cols},
                                            **{'text': original_text}
                                            })
                    fig = px.scatter(tsne_df, x="x", y="y", color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Sentence-Embeddings =`{e_com_storage_ref}`, Manifold-Algo =`{algo}` for `D=2`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig, key=key)
                    col_index += 1
                    if are_cols_full():
                        cols = st.columns(num_cols)
                        col_index = 0
                if 3 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo, 3, n_jobs).fit_transform(mat)
                    x = low_dim_data[:, 0]
                    y = low_dim_data[:, 1]
                    z = low_dim_data[:, 2]
                    tsne_df = pd.DataFrame({**{'x': x, 'y': y, 'z': z},
                                            **{k: predictions[k] for k in classifier_cols},
                                            **{'text': original_text}
                                            })
                    fig = px.scatter_3d(tsne_df, x="x", y="y", z='z', color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Sentence-Embeddings =`{e_com_storage_ref}`, Manifold-Algo =`{algo}` for `D=3`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig, key=key)
                    col_index += 1
                    if are_cols_full():
                        cols = st.columns(num_cols)
                        col_index = 0

            # Todo fancy embed infos etc
            # if display_embed_information: display_embed_vetor_information(e_com,mat)

        # if display_embed_information:
        #     exp = st.expander("Embedding vector information")
        #     exp.write(embed_vector_info)

        if show_infos:
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes=[pipe])
            StreamlitVizTracker.display_footer()

# vodafonegmbh 40875 Radtingen
